@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2008-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2008-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -73,6 +73,7 @@ typedef struct ContextDma ContextDma;
                 {                                                                                  \
                     root_alloc_params.processID = pClient->ProcID;                                 \
                     NV_ASSERT(root_alloc_params.processID == osGetCurrentProcess());               \
+                    root_alloc_params.pOsPidInfo = (NvP64)(pClient->pOsPidInfo);                   \
                 }                                                                                  \
             }                                                                                      \
             else                                                                                   \
@@ -202,7 +203,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                              \
         if ((status == NV_OK) && (pRpc != NULL))                                              \
         {                                                                                     \
-            if (IS_GSP_CLIENT(pGpu))                                                          \
+            if (IS_FW_CLIENT(pGpu))                                                           \
             {                                                                                 \
                 NV_RM_RPC_ALLOC_SHARE_DEVICE_FWCLIENT(pGpu, hclient, hdevice, hclientshare, htargetclient, htargetdevice, hclass, \
                                                       allocflags, vasize, vamode, bFirstDevice, status); \
@@ -226,7 +227,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                    \
         if ((status == NV_OK) && (pRpc != NULL))                                    \
         {                                                                           \
-            if (IS_GSP_CLIENT(pGpu))                                                \
+            if (IS_FW_CLIENT(pGpu))                                                 \
             {                                                                       \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                      \
                 status = pRmApi->Control(pRmApi, hClient, hObject, cmd,             \
@@ -250,7 +251,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                        \
         if ((status == NV_OK) && (pRpc != NULL))                                        \
         {                                                                               \
-            if (IS_GSP_CLIENT(pGpu))                                                    \
+            if (IS_FW_CLIENT(pGpu))                                                     \
             {                                                                           \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                          \
                 status = pRmApi->AllocWithHandle(pRmApi, hclient, hparent, hchannel,    \
@@ -274,7 +275,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                        \
         if ((status == NV_OK) && (pRpc != NULL))                                        \
         {                                                                               \
-            if (IS_GSP_CLIENT(pGpu))                                                    \
+            if (IS_FW_CLIENT(pGpu))                                                     \
             {                                                                           \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                          \
                 status = pRmApi->AllocWithHandle(pRmApi, hclient, hchannel, hobject,    \
@@ -298,7 +299,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                        \
         if ((status == NV_OK) && (pRpc != NULL))                        \
         {                                                               \
-            if (IS_GSP_CLIENT(pGpu))                                    \
+            if (IS_FW_CLIENT(pGpu))                                     \
             {                                                           \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);          \
                 status = pRmApi->Free(pRmApi, hclient, hobject);        \
@@ -319,7 +320,7 @@ typedef struct ContextDma ContextDma;
         /* used in failure cases, macro doesn't overwrite rmStatus */   \
         if (pRpc != NULL)                                               \
         {                                                               \
-            if (IS_GSP_CLIENT(pGpu))                                    \
+            if (IS_FW_CLIENT(pGpu))                                     \
             {                                                           \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);          \
                 pRmApi->Free(pRmApi, hclient, hobject);                 \
@@ -341,7 +342,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                   \
         if ((status == NV_OK) && (pRpc != NULL))                                   \
         {                                                                          \
-            if (IS_GSP_CLIENT(pGpu))                                               \
+            if (IS_FW_CLIENT(pGpu))                                                \
             {                                                                      \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                     \
                 NV0005_ALLOC_PARAMETERS allocParams = {0};                         \
@@ -371,7 +372,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                        \
         if ((status == NV_OK) && (pRpc != NULL))                                        \
         {                                                                               \
-            if (IS_GSP_CLIENT(pGpu))                                                    \
+            if (IS_FW_CLIENT(pGpu))                                                     \
             {                                                                           \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                          \
                 NV2080_ALLOC_PARAMETERS alloc_params = {0};                             \
@@ -397,7 +398,7 @@ typedef struct ContextDma ContextDma;
         NV_ASSERT(pRpc != NULL);                                                \
         if ((status == NV_OK) && (pRpc != NULL))                                \
         {                                                                       \
-            if (IS_GSP_CLIENT(pGpu))                                            \
+            if (IS_FW_CLIENT(pGpu))                                             \
             {                                                                   \
                 RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                  \
                 status = pRmApi->DupObject(pRmApi, hclient, hparent,            \
@@ -436,17 +437,6 @@ static inline void NV_RM_RPC_SIM_FREE_INFRA(OBJGPU *pGpu, ...) { return; }
             status = NV_ERR_INSUFFICIENT_RESOURCES;                     \
     } while(0)
 
-#define NV_RM_RPC_SET_SYSMEM_DIRTY_PAGE_TRACKING_BUFFER(pGpu, status, bitmapInfo)    \
-    do                                                                               \
-    {                                                                                \
-        OBJRPC *pRpc = GPU_GET_RPC(pGpu);                                            \
-        NV_ASSERT(pRpc != NULL);                                                     \
-        if ((status == NV_OK) && (pRpc != NULL))                                     \
-            status = rpcSetSysmemDirtyPageTrackingBuffer_HAL(pGpu, pRpc, bitmapInfo);\
-        else if (pRpc == NULL)                                                       \
-            status = NV_ERR_INSUFFICIENT_RESOURCES;                                  \
-    } while(0)
-
 #define NV_RM_RPC_UNLOADING_GUEST_DRIVER(pGpu, status, bInPMTransition, bGc6Entering, newPMLevel)        \
     do                                                                                            \
     {                                                                                             \
@@ -493,28 +483,6 @@ static inline void NV_RM_RPC_SIM_FREE_INFRA(OBJGPU *pGpu, ...) { return; }
             status = NV_ERR_INSUFFICIENT_RESOURCES;                                         \
     } while (0)
 
-#define NV_RM_RPC_SET_PAGE_DIRECTORY(pGpu, hClient, hDevice, pParams, status) \
-    do                                                                        \
-    {                                                                         \
-        OBJRPC *pRpc = GPU_GET_RPC(pGpu);                                     \
-        NV_ASSERT(pRpc != NULL);                                              \
-        if ((status == NV_OK) && (pRpc != NULL))                              \
-            status = rpcSetPageDirectory_HAL(pGpu, pRpc, hClient, hDevice, pParams);  \
-        else if (pRpc == NULL)                                                 \
-            status = NV_ERR_INSUFFICIENT_RESOURCES;                            \
-    } while (0)
-
-#define NV_RM_RPC_UNSET_PAGE_DIRECTORY(pGpu, hClient, hDevice, pParams, status) \
-    do                                                                          \
-    {                                                                           \
-        OBJRPC *pRpc = GPU_GET_RPC(pGpu);                                       \
-        NV_ASSERT(pRpc != NULL);                                                \
-        if ((status == NV_OK) && (pRpc != NULL))                                \
-            status = rpcUnsetPageDirectory_HAL(pGpu, pRpc, hClient, hDevice, pParams);  \
-        else if (pRpc == NULL)                                                   \
-            status = NV_ERR_INSUFFICIENT_RESOURCES;                              \
-    } while (0)
-
 static inline void NV_RM_RPC_PMA_SCRUBBER_SHARED_BUFFER_GUEST_PAGES_OPERATION(OBJGPU *pGpu, ...) { return; }
 
 #define NV_RM_RPC_INVALIDATE_TLB(pGpu, status, pdbAddress, regVal)         \
@@ -532,7 +500,17 @@ static inline void NV_RM_RPC_PMA_SCRUBBER_SHARED_BUFFER_GUEST_PAGES_OPERATION(OB
 // DCE_CLIENT_RM specific RPCs
 //
 
-#define NV_RM_RPC_DCE_RM_INIT(pGpu, bInit, status)      do {} while (0)
+#define NV_RM_RPC_DCE_RM_INIT(pGpu, bInit, status)                              \
+    do                                                                          \
+    {                                                                           \
+        OBJRPC* pRpc = GPU_GET_RPC(pGpu);                                       \
+        if ((status == NV_OK) && (pRpc != NULL))                                \
+        {                                                                       \
+            RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);                      \
+            status = rpcDceRmInit_dce(pRmApi, bInit);                           \
+        } else if (pRpc == NULL)                                                \
+            status = NV_ERR_INSUFFICIENT_RESOURCES;                             \
+    } while (0)
 
 //
 // GSP_CLIENT_RM specific RPCs
