@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1993-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -36,6 +36,7 @@
 #include "nvlink_inband_msg.h"
 #include "kernel/mem_mgr/fabric_vaspace.h"
 #include "ctrl/ctrl2080/ctrl2080internal.h"
+#include "nvrm_registry.h"
 
 #include "compute/fabric.h"
 #include "class/cl00f1.h"
@@ -80,6 +81,45 @@ _gpuFabricProbeFullSanityCheck
     return NV_OK;
 }
 
+static void
+_gpuFabricProbeInvalidate
+(
+    OBJGPU *pGpu
+)
+{
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel =
+                                    pGpu->pGpuFabricProbeInfoKernel;
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    FABRIC_VASPACE *pFabricVAS = dynamicCast(pGpu->pFabricVAS, FABRIC_VASPACE);
+
+    portAtomicSetU32(&pGpuFabricProbeInfoKernel->probeRespRcvd, 0);
+
+    if (pKernelNvlink != NULL)
+    {
+        knvlinkClearUniqueFabricBaseAddress_HAL(pGpu, pKernelNvlink);
+        knvlinkClearUniqueFabricEgmBaseAddress_HAL(pGpu, pKernelNvlink);
+    }
+
+    if (pFabricVAS != NULL)
+        fabricvaspaceClearUCRange(pFabricVAS);
+}
+
+static void
+_gpuFabricProbeCheckResetRequired
+(
+    OBJGPU *pGpu,
+    nvlink_inband_gpu_probe_rsp_msg_t *pProbeRespMsg
+)
+{
+    if (REF_VAL(NVLINK_INBAND_FABRIC_HEALTH_MASK_INCORRECT_CONFIGURATION, pProbeRespMsg->probeRsp.fabricHealthMask) ==
+        NVLINK_INBAND_FABRIC_HEALTH_MASK_INCORRECT_CONFIGURATION_GPU_STATE_INVALID)
+    {
+        pGpu->setProperty(pGpu, PDB_PROP_GPU_RECOVERY_SQUASH_XID154, NV_TRUE);
+        gpuMarkDeviceForDrainAndReset(pGpu);
+        pGpu->setProperty(pGpu, PDB_PROP_GPU_RECOVERY_SQUASH_XID154, NV_FALSE);
+    }
+}
+
 NV_STATUS
 gpuFabricProbeGetGpuFabricHandle
 (
@@ -245,6 +285,28 @@ gpuFabricProbeGetFlaAddressRange
     return status;
 }
 
+/*
+ * This function is used to get the peer GPU EGM address from FM to RM.
+ * FM passes only the upper 32 bits of the address.
+ */
+NV_STATUS
+gpuFabricProbeGetEgmGpaAddress
+(
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel,
+    NvU64 *pEgmGpaAddress
+)
+{
+    NV_STATUS status;
+
+    status = _gpuFabricProbeFullSanityCheck(pGpuFabricProbeInfoKernel);
+
+    NV_CHECK_OR_RETURN(LEVEL_SILENT, status == NV_OK, status);
+
+    *pEgmGpaAddress = (NvU64)pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.gpaAddressEGMHi << 32ULL;
+
+    return status;
+}
+
 NV_STATUS
 gpuFabricProbeGetNumProbeReqs
 (
@@ -264,8 +326,9 @@ gpuFabricProbeGetNumProbeReqs
     pGpu = pGpuFabricProbeInfoKernel->pGpu;
     pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
 
-    LOCK_ASSERT_AND_RETURN(rmDeviceGpuLockIsOwner(
-                           gpuGetInstance(pGpuFabricProbeInfoKernel->pGpu)));
+    NV_ASSERT_OR_RETURN(rmDeviceGpuLockIsOwner(
+                           gpuGetInstance(pGpuFabricProbeInfoKernel->pGpu)),
+        NV_ERR_INVALID_LOCK_STATE);
 
     //
     // This function gets called from CUDA stack to check the probe state (NV2080_CTRL_CMD_GET_GPU_FABRIC_PROBE_INFO).
@@ -390,6 +453,8 @@ _gpuFabricProbeSetupGpaRange
     {
         NvU64 gpaAddress;
         NvU64 gpaAddressSize;
+        NvU64 egmGpaAddress;
+        NvU64 fmCaps;
 
         NV_CHECK_OR_RETURN_VOID(LEVEL_ERROR,
                     gpuFabricProbeGetGpaAddress(pGpuFabricProbeInfoKernel,
@@ -402,6 +467,21 @@ _gpuFabricProbeSetupGpaRange
         NV_CHECK_OR_RETURN_VOID(LEVEL_ERROR,
                     knvlinkSetUniqueFabricBaseAddress_HAL(pGpu, pKernelNvlink,
                                                         gpaAddress) == NV_OK);
+
+        NV_CHECK_OR_RETURN_VOID(LEVEL_ERROR,
+                    gpuFabricProbeGetfmCaps(pGpuFabricProbeInfoKernel,
+                                        &fmCaps) == NV_OK);
+
+        if (fmCaps & NVLINK_INBAND_FM_CAPS_EGM_ENABLED)
+        {
+            NV_CHECK_OR_RETURN_VOID(LEVEL_ERROR,
+                        gpuFabricProbeGetEgmGpaAddress(pGpuFabricProbeInfoKernel,
+                                                    &egmGpaAddress) == NV_OK);
+
+            NV_CHECK_OR_RETURN_VOID(LEVEL_ERROR,
+                    knvlinkSetUniqueFabricEgmBaseAddress_HAL(pGpu, pKernelNvlink,
+                                                        egmGpaAddress) == NV_OK);
+        }
     }
 }
 
@@ -459,6 +539,59 @@ _gpuFabricProbeSendCliqueIdChangeEvent
     }
 }
 
+static void
+_gpuFabricProbeUpdateSupportedBwModes
+(
+    OBJGPU *pGpu,
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel
+)
+{
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NvU8 maxRbmLinks = pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.maxRbmLinks;
+
+    if (pKernelNvlink != NULL)
+    {
+        // Check if maxRbmLinks is within max supported links
+        NV_ASSERT_OR_RETURN_VOID(maxRbmLinks <= NVLINK_MAX_LINKS_SW);
+        knvlinkSetMaxBWModeLinks(pGpu, pKernelNvlink, maxRbmLinks);
+    }
+}
+
+static void
+_gpuFabricProbeRbmSleepLinks
+(
+    OBJGPU *pGpu,
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel
+)
+{
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NV_STATUS status = NV_OK;
+    NvU32 enabledLinkMask;
+    NvU32 linkMask;
+    NvBool bEnableABM = knvlinkGetAbmEnabled(pGpu, pKernelNvlink);
+
+    // Bail if RBM was not requested and ABM is not enabled/advertised
+    if ((DRF_VAL(_GPU, _NVLINK, _BW_MODE, pGpuFabricProbeInfoKernel->bwMode) != GPU_NVLINK_BW_MODE_LINK_COUNT) &&
+        (!bEnableABM ||
+         !(pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.fmCaps & NVLINK_INBAND_GPU_PROBE_CAPS_ADAPTIVE_BANDWIDTH_SUPPORT)))
+    {
+        return;
+    }
+
+    // Set links requested in linkMaskToBeReduced to sleep
+    enabledLinkMask = knvlinkGetEnabledLinkMask(pGpu, pKernelNvlink);
+    linkMask = pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.linkMaskToBeReduced &
+                            enabledLinkMask;
+    NV_PRINTF(LEVEL_NOTICE, "GPU%u Updating RBM/ABM linkmask via probe request: linkMask=0x%x\n",gpuGetInstance(pGpu), linkMask);
+
+    status = knvlinkEnterExitSleep(pGpu, pKernelNvlink, linkMask, NV_TRUE);
+    if (status != NV_OK)
+    {
+        NV_PRINTF(LEVEL_ERROR, "Error setting links to sleep on linkmask 0x%x\n",
+                  linkMask);
+    }
+}
+
 NV_STATUS
 gpuFabricProbeReceiveKernelCallback
 (
@@ -471,6 +604,7 @@ gpuFabricProbeReceiveKernelCallback
     NvU32 gpuMaskUnused;
     nvlink_inband_gpu_probe_rsp_msg_t *pProbeRespMsg;
     GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel;
+    KernelNvlink *pKernelNvlink;
     NV_STATUS status;
 
     if ((pGpu = gpumgrGetGpu(gpuInstance)) == NULL)
@@ -513,14 +647,59 @@ gpuFabricProbeReceiveKernelCallback
     portAtomicSetU32(&pGpuFabricProbeInfoKernel->probeRespRcvd, 1);
 
     status = _gpuFabricProbeFullSanityCheck(pGpuFabricProbeInfoKernel);
-    NV_CHECK_OR_RETURN(LEVEL_ERROR, status == NV_OK, status);
+    NV_CHECK_OR_RETURN(LEVEL_INFO, status == NV_OK, status);
 
     _gpuFabricProbeSetupGpaRange(pGpu, pGpuFabricProbeInfoKernel);
     _gpuFabricProbeSetupFlaRange(pGpu, pGpuFabricProbeInfoKernel);
 
+    // Update supported bandwidth modes from probe response
+    _gpuFabricProbeUpdateSupportedBwModes(pGpu, pGpuFabricProbeInfoKernel);
+
+    _gpuFabricProbeCheckResetRequired(pGpu, pProbeRespMsg);
+
+    pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    if (pKernelNvlink != NULL)
+    {
+        if (pKernelNvlink->getProperty(pKernelNvlink, PDB_PROP_KNVLINK_RBM_LINK_COUNT_ENABLED))
+        {
+            // Set links to sleep based on probe response
+            _gpuFabricProbeRbmSleepLinks(pGpu, pGpuFabricProbeInfoKernel);
+        }
+    }
+
+
     return NV_OK;
 }
 
+static void
+_gpuFabricProbeUpdateABMLinkMask
+(
+    OBJGPU *pGpu,
+    NvU8 action,
+    NvU64 linkMask
+)
+{
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NvBool bNeedsRCRecovery = NV_FALSE;
+    NvU64 enabledLinkMask;
+
+    if (pKernelNvlink == NULL)
+        return;
+
+    if (action != NVLINK_INBAND_GPU_PROBE_UPDATE_ACTION_CHANGE_AMAP &&
+        action != NVLINK_INBAND_GPU_PROBE_UPDATE_ACTION_CHANGE_AMAP_AND_QUIESCE)
+        return;
+
+    // Store linkmask to be updated after P2P is idle
+    enabledLinkMask = knvlinkGetEnabledLinkMask(pGpu, pKernelNvlink);
+    knvlinkSetPendingAbmLinkMaskToBeReduced(pGpu, pKernelNvlink,(NvU32)(enabledLinkMask & ~linkMask));
+
+    NV_PRINTF(LEVEL_NOTICE, "GPU%u Queing ABM linkmask via probe update: linkMask=0x%llx\n", pGpu->gpuInstance, linkMask);
+
+    bNeedsRCRecovery = (action == NVLINK_INBAND_GPU_PROBE_UPDATE_ACTION_CHANGE_AMAP_AND_QUIESCE);
+    knvlinkABMLinkMaskUpdate(pGpu, pKernelNvlink, bNeedsRCRecovery);
+}
+
 NV_STATUS
 gpuFabricProbeReceiveUpdateKernelCallback
 (
@@ -534,6 +713,8 @@ gpuFabricProbeReceiveUpdateKernelCallback
     nvlink_inband_gpu_probe_update_req_msg_t *pProbeUpdateReqMsg;
     nvlink_inband_gpu_probe_rsp_msg_t *pProbeRespMsg;
     GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel;
+    KernelNvlink *pKernelNvlink;
+    NvBool bEnableABM;
     NV_STATUS status;
 
     if ((pGpu = gpumgrGetGpu(gpuInstance)) == NULL)
@@ -541,6 +722,8 @@ gpuFabricProbeReceiveUpdateKernelCallback
         NV_ASSERT_FAILED("Invalid GPU instance");
         return NV_ERR_INVALID_ARGUMENT;
     }
+    pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    bEnableABM = knvlinkGetAbmEnabled(pGpu, pKernelNvlink);
 
     NV_CHECK_OR_RETURN(LEVEL_ERROR, pGpu->pGpuFabricProbeInfoKernel != NULL, NV_OK);
 
@@ -571,6 +754,25 @@ gpuFabricProbeReceiveUpdateKernelCallback
                                                pProbeRespMsg->probeRsp.cliqueId);
     }
 
+     // Support for fabric attributes enhancements (Bug: 5345385)
+    if (pProbeUpdateReqMsg->probeUpdate.action == NVLINK_INBAND_GPU_PROBE_UPDATE_ACTION_PROBE_REQUEST_NEEDED)
+    {
+        pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+        if (pKernelNvlink != NULL)
+        {
+            knvlinkTriggerProbeRequest(pGpu, pKernelNvlink);
+        }
+    }
+
+    _gpuFabricProbeCheckResetRequired(pGpu, pProbeRespMsg);
+
+    if (bEnableABM)
+    {
+        _gpuFabricProbeUpdateABMLinkMask(pGpu,
+                                         pProbeUpdateReqMsg->probeUpdate.action,
+                                         pProbeUpdateReqMsg->probeUpdate.linkMask);
+    }
+
     return NV_OK;
 }
 
@@ -602,6 +804,36 @@ gpuFabricProbeSuspend
                             NULL, 0));
 }
 
+void
+gpuFabricProbeInvalidate
+(
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel
+)
+{
+    OBJGPU *pGpu;
+    RM_API *pRmApi;
+    NV_STATUS status;
+
+    if (pGpuFabricProbeInfoKernel == NULL)
+    {
+        return;
+    }
+
+    pGpu = pGpuFabricProbeInfoKernel->pGpu;
+    pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
+
+    NV_ASSERT(rmDeviceGpuLockIsOwner(gpuGetInstance(pGpu)));
+
+    NV_CHECK_OK(status, LEVEL_ERROR,
+            pRmApi->Control(pRmApi,
+                            pGpu->hInternalClient,
+                            pGpu->hInternalSubdevice,
+                            NV2080_CTRL_CMD_INTERNAL_GPU_INVALIDATE_FABRIC_PROBE,
+                            NULL, 0));
+
+    _gpuFabricProbeInvalidate(pGpu);
+}
+
 NV_STATUS
 gpuFabricProbeResume
 (
@@ -634,6 +866,49 @@ gpuFabricProbeResume
     return NV_OK;
 }
 
+static void
+_gpuFabricProbeRbmWakeLinks
+(
+    OBJGPU *pGpu,
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel
+)
+{
+    RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NV2080_CTRL_NVLINK_GET_POWER_STATE_PARAMS powerStatusParams = {0};
+    NvU32 enabledLinkMask = knvlinkGetEnabledLinkMask(pGpu, pKernelNvlink);
+    NvU32 linkMask = 0;
+    NV_STATUS status;
+    NvU32 i;
+
+    FOR_EACH_INDEX_IN_MASK(32, i, enabledLinkMask)
+    {
+        powerStatusParams.linkId = i;
+        NV_CHECK_OK(status, LEVEL_INFO,
+            pRmApi->Control(pRmApi,
+                            pGpu->hInternalClient,
+                            pGpu->hInternalSubdevice,
+                            NV2080_CTRL_CMD_NVLINK_GET_POWER_STATE,
+                            &powerStatusParams, sizeof(powerStatusParams)));
+        if (powerStatusParams.powerState == NV2080_CTRL_NVLINK_POWER_STATE_L2)
+        {
+            linkMask |= NVBIT32(i);
+        }
+    }
+    FOR_EACH_INDEX_IN_MASK_END;
+
+    if (linkMask)
+    {
+        // Wake all sleeping links
+        status = knvlinkEnterExitSleep(pGpu, pKernelNvlink, linkMask, NV_FALSE);
+        if (status != NV_OK)
+        {
+            NV_PRINTF(LEVEL_ERROR, "Error waking links on linkmask 0x%x\n",
+                      linkMask);
+        }
+    }
+}
+
 NV_STATUS
 gpuFabricProbeStart
 (
@@ -641,12 +916,20 @@ gpuFabricProbeStart
     GPU_FABRIC_PROBE_INFO_KERNEL **ppGpuFabricProbeInfoKernel
 )
 {
+    OBJSYS *pSys = SYS_GET_INSTANCE();
+    OBJGPUMGR *pGpuMgr = SYS_GET_GPUMGR(pSys);
     NV_STATUS status = NV_OK;
     GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel;
     RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
     NV2080_CTRL_CMD_INTERNAL_START_GPU_FABRIC_PROBE_INFO_PARAMS params = { 0 };
+    MemoryManager *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NvU32 strLength = 32;
+    NvU8 pStr[32];
+    NvU32 linkCount;
 
-    LOCK_ASSERT_AND_RETURN(rmDeviceGpuLockIsOwner(gpuGetInstance(pGpu)));
+    NV_ASSERT_OR_RETURN(rmDeviceGpuLockIsOwner(gpuGetInstance(pGpu)), 
+        NV_ERR_INVALID_LOCK_STATE);
 
     // Check if NVSwitch based system. If not return without doing anything
     if (!gpuFabricProbeIsSupported(pGpu))
@@ -663,9 +946,35 @@ gpuFabricProbeStart
     portMemSet(pGpuFabricProbeInfoKernel, 0, sizeof(*pGpuFabricProbeInfoKernel));
 
     pGpuFabricProbeInfoKernel->pGpu = pGpu;
-    pGpuFabricProbeInfoKernel->bwMode = gpumgrGetGpuNvlinkBwMode();
+    params.bLocalEgmEnabled = pMemoryManager->bLocalEgmEnabled;
+
+    if (pGpuMgr->bwModeScope == GPU_NVLINK_BW_MODE_SCOPE_PER_GPU)
+    {
+        if (pKernelNvlink != NULL)
+        {
+            pGpuFabricProbeInfoKernel->bwMode = knvlinkGetBWMode(pGpu, pKernelNvlink);
+        }
+    }
+    else
+    {
+        pGpuFabricProbeInfoKernel->bwMode = gpumgrGetGpuNvlinkBwMode();
+    }
     params.bwMode = pGpuFabricProbeInfoKernel->bwMode;
 
+    // pKernelNvlink is allowed to be NULL for vGPU case to send probe via RPC. Skip RBM for this case.
+    if (pKernelNvlink != NULL)
+    {
+        // RBM requested via regkey. Wake the links in this path if LINKCOUNT or FULL is set.
+        if ((osReadRegistryDword(pGpu, NV_REG_STR_RM_NVLINK_BW_LINK_COUNT, &linkCount) == NV_OK) ||
+            ((osReadRegistryString(pGpu, NV_REG_STR_RM_NVLINK_BW, pStr, &strLength) == NV_OK) &&
+            (DRF_VAL(_GPU, _NVLINK, _BW_MODE, params.bwMode) == GPU_NVLINK_BW_MODE_FULL)))
+        {
+            if (pKernelNvlink->getProperty(pKernelNvlink, PDB_PROP_KNVLINK_RBM_LINK_COUNT_ENABLED))
+            {
+                _gpuFabricProbeRbmWakeLinks(pGpu, pGpuFabricProbeInfoKernel);
+            }
+        }
+    }
 
     if (IS_VIRTUAL(pGpu))
     {
@@ -760,26 +1069,6 @@ gpuFabricProbeIsSupported
     return NV_TRUE;
 }
 
-static void
-_gpuFabricProbeInvalidate
-(
-    OBJGPU *pGpu
-)
-{
-    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel =
-                                    pGpu->pGpuFabricProbeInfoKernel;
-    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
-    FABRIC_VASPACE *pFabricVAS = dynamicCast(pGpu->pFabricVAS, FABRIC_VASPACE);
-
-    portAtomicSetU32(&pGpuFabricProbeInfoKernel->probeRespRcvd, 0);
-
-    if (pKernelNvlink != NULL)
-        knvlinkClearUniqueFabricBaseAddress_HAL(pGpu, pKernelNvlink);
-
-    if (pFabricVAS != NULL)
-        fabricvaspaceClearUCRange(pFabricVAS);
-}
-
 #define GPU_FABRIC_CHECK_BW_MODE(fmCaps, mode)                    \
     do                                                            \
     {                                                             \
@@ -787,17 +1076,19 @@ _gpuFabricProbeInvalidate
             return NV_ERR_NOT_SUPPORTED;                          \
     } while (0)
 
-static NV_STATUS
-_gpuFabricProbeUpdateBwMode
+NV_STATUS
+gpuFabricProbeSetBwModePerGpu
 (
     OBJGPU *pGpu,
     NvU8 mode
 )
 {
     GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel =
-                                                pGpu->pGpuFabricProbeInfoKernel;
+                                    pGpu->pGpuFabricProbeInfoKernel;
     RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
+    KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
 
+    // Function assumes caller has checked requested mode is supported
     pGpuFabricProbeInfoKernel->bwMode = mode;
 
     gpuFabricProbeSuspend(pGpuFabricProbeInfoKernel);
@@ -811,6 +1102,17 @@ _gpuFabricProbeUpdateBwMode
 
     _gpuFabricProbeInvalidate(pGpu);
 
+    // Need to wake links if LINK_COUNT or FULL BW mode is requested before probe update
+    if ((DRF_VAL(_GPU, _NVLINK, _BW_MODE, mode) == GPU_NVLINK_BW_MODE_LINK_COUNT) ||
+        (DRF_VAL(_GPU, _NVLINK, _BW_MODE, mode) == GPU_NVLINK_BW_MODE_FULL))
+    {
+        if (pKernelNvlink->getProperty(pKernelNvlink, PDB_PROP_KNVLINK_RBM_LINK_COUNT_ENABLED))
+        {
+            _gpuFabricProbeRbmWakeLinks(pGpu, pGpuFabricProbeInfoKernel);
+            knvlinkSetBWMode(pGpu, pKernelNvlink, mode);
+        }
+    }
+
     NV_CHECK_OK_OR_RETURN(LEVEL_ERROR, gpuFabricProbeResume(pGpuFabricProbeInfoKernel));
 
     return NV_OK;
@@ -827,6 +1129,7 @@ gpuFabricProbeSetBwMode
     NV_STATUS status;
     NvU32 gpuIndex;
     OBJGPU *pGpu;
+    KernelNvlink *pKernelNvlink;
 
     status = gpumgrGetGpuAttachInfo(&attachedGpuCount, &attachedGpuMask);
     if (status != NV_OK)
@@ -862,22 +1165,34 @@ gpuFabricProbeSetBwMode
             return NV_ERR_NOT_READY;
         }
 
-        NvU64 fmCaps = pGpu->pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.fmCaps;
-        switch(mode)
+        // Confirm requested BW mode is supported by GPU
+        pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+        if (!knvlinkIsBwModeSupported_HAL(pGpu, pKernelNvlink, mode))
         {
-            case GPU_NVLINK_BW_MODE_MIN:
-                GPU_FABRIC_CHECK_BW_MODE(fmCaps, MIN);
-                break;
-            case GPU_NVLINK_BW_MODE_HALF:
-                GPU_FABRIC_CHECK_BW_MODE(fmCaps, HALF);
-                break;
-            case GPU_NVLINK_BW_MODE_3QUARTER:
-                GPU_FABRIC_CHECK_BW_MODE(fmCaps, 3QUARTER);
-                break;
-            case GPU_NVLINK_BW_MODE_OFF:
-                return NV_OK; // Don't need to ask FM
-            default:
-                break;
+            return NV_ERR_NOT_SUPPORTED;
+        }
+
+        // These checks only apply to legacy bw mode settings
+        if (DRF_VAL(_GPU, _NVLINK, _BW_MODE, mode) !=
+            GPU_NVLINK_BW_MODE_LINK_COUNT)
+        {
+            NvU64 fmCaps = pGpu->pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.fmCaps;
+            switch(mode)
+            {
+                case GPU_NVLINK_BW_MODE_MIN:
+                    GPU_FABRIC_CHECK_BW_MODE(fmCaps, MIN);
+                    break;
+                case GPU_NVLINK_BW_MODE_HALF:
+                    GPU_FABRIC_CHECK_BW_MODE(fmCaps, HALF);
+                    break;
+                case GPU_NVLINK_BW_MODE_3QUARTER:
+                    GPU_FABRIC_CHECK_BW_MODE(fmCaps, 3QUARTER);
+                    break;
+                case GPU_NVLINK_BW_MODE_OFF:
+                    return NV_OK; // Don't need to ask FM
+                default:
+                    break;
+            }
         }
     }
 
@@ -904,7 +1219,7 @@ gpuFabricProbeSetBwMode
         pGpu != NULL;
         pGpu = gpumgrGetNextGpu(attachedGpuMask, &gpuIndex))
     {
-        status = _gpuFabricProbeUpdateBwMode(pGpu, mode);
+        status = gpuFabricProbeSetBwModePerGpu(pGpu, mode);
         if (status != NV_OK)
         {
             return status;
@@ -932,3 +1247,41 @@ gpuFabricProbeGetlinkMaskToBeReduced
 
     return NV_OK;
 }
+
+NV_STATUS
+gpuFabricProbeSetlinkMaskToBeReduced
+(
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel,
+    NvU32 linkMaskToBeReduced
+)
+{
+    NV_STATUS status;
+
+    status = _gpuFabricProbeFullSanityCheck(pGpuFabricProbeInfoKernel);
+
+    NV_CHECK_OR_RETURN(LEVEL_SILENT, status == NV_OK, status);
+
+    pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.linkMaskToBeReduced =
+        linkMaskToBeReduced;
+
+    return NV_OK;
+}
+
+NV_STATUS
+gpuFabricProbeGetRemapTableIndex
+(
+    GPU_FABRIC_PROBE_INFO_KERNEL *pGpuFabricProbeInfoKernel,
+    NvU32 *pRemapTableIdx
+)
+{
+    NV_STATUS status;
+
+    status = _gpuFabricProbeFullSanityCheck(pGpuFabricProbeInfoKernel);
+
+    NV_CHECK_OR_RETURN(LEVEL_SILENT, status == NV_OK, status);
+
+    *pRemapTableIdx =
+        pGpuFabricProbeInfoKernel->probeResponseMsg.probeRsp.remapTableIdx;
+
+    return NV_OK;
+}
